package bitRanks

import ap.parser.{Environment, IFormula, Parser2InputAbsy}
import ap.terfor.ConstantTerm

import scala.util.parsing.combinator.RegexParsers

object TransitionParser extends RegexParsers {

  class TransitionParserException(msg : String) extends Exception(msg)
  
  def apply(reader : java.io.Reader, env : Environment) 
                    : (List[ConstantTerm], List[ConstantTerm], IFormula) = {
    val parseRes = parse(transitionDef, reader)

    parseRes match {
      case Success((from, to, transition), _) => {
        if (from.size != to.size)
          throw new TransitionParserException("Different numbers of input and output variables")

        val fromConsts = for (name <- from) yield new ConstantTerm(name)
        val toConsts = for (name <- to) yield new ConstantTerm(name)
        
        for (c <- fromConsts ++ toConsts)
          env.addConstant(c, Environment.Existential)
        
        val expr =
          Parser2InputAbsy.parseExpression(new java.io.StringReader(transition), env)
        
        if (!expr.isInstanceOf[IFormula])
          throw new TransitionParserException("Transition relation has to be a formula")
          
        (fromConsts, toConsts, expr.asInstanceOf[IFormula])
      }
      case res : NoSuccess =>
        throw new TransitionParserException(res.msg)
    }
  }
    
  /**
   * Parse a single triple consisting of <code>\from</code>, <code>\to</code>,
   * and <code>\transition</code>. This also gets rid of comments and passes
   * the blocks to the individual parsers. We read the input line by line
   */
/*  private class ChunkParser(reader : java.io.Reader) {
    
    private val lineReader = new java.io.BufferedReader(reader)
    
    private val lineIt = new Iterator[String] {
      private var nextLine = lineReader.readLine
      def hasNext = (nextLine != null)
      def next = {
        val res = nextLine
        nextLine = lineReader.readLine
        res
      }
    }
    
    private val lineCommentLessIt : Iterator[String] =
      for (l <- lineIt) yield lineComment.replaceAllIn(l, "")
    
    private val commentLessIt : Iterator[String] = {
      
      for (l <- lineCommentLessIt) yield {
        
      }
    }
      
    
    private val charIt : Iterator[Char] =
      for (l <- lineCommentLessIt; c <- l.elements) yield c
    
  }
   */
//  private val lineComment = """//.*$""".r
//  private val commentStart = """/\*""".r
//  private val commentEnd = """\*/""".r


  //////////////////////////////////////////////////////////////////////////////
  // The actual parser (which considers the transition relation simply as a
  // (correctly nested) string that is later passed to the Princess parser)
  
  private val ident : Parser[String] = """[a-zA-Z]([a-zA-Z]|[0-9]|_|')*""".r
    
  private def varList : Parser[List[String]] =
    "{" ~> (rep1sep(ident, ";") <~ ";" | success(List())) <~ "}"

  private def fromList : Parser[List[String]] = "\\from" ~> varList
  private def toList : Parser[List[String]] =   "\\to" ~> varList

  private val bracelessString : Parser[String] = """[^{}]*""".r
  private def correctlyNestedString : Parser[String] =
    (bracelessString ~
     (("{" ~ correctlyNestedString ~ "}" ~ correctlyNestedString) ^^
        {case "{"~x~"}"~y => "{"+x+"}"+y} |
      success(""))) ^^
    {case x~y => x+y}
  
  private def transition : Parser[String] =
    "\\transition" ~ "{" ~> correctlyNestedString <~ "}"
  
  private def transitionDef : Parser[(List[String], List[String], String)] =
    (fromList ~ toList ~ transition) ^^ {case x~y~z => (x, y, z)}
  
}
